/*
 * File Name:sm2_phytium.c - Phytium SDK for SM2
 *
 * Copyright (C) 2020 Phytium Technology Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */
#include <linux/scatterlist.h>
#include <asm/memory.h>
#include <linux/delay.h>
#include <crypto/internal/akcipher.h>
#include <crypto/akcipher.h>
#include <linux/io.h>
#include <linux/crypto.h>
#include <linux/kthread.h>
#include <asm/cacheflush.h>
#include "smx_common.h"
#include "sm2_phytium.h"
#include "sm3_phytium.h"
#include "phytium_scto.h"

#if SCTO_KERNEL_MODE
static void sm3_init(struct sm3_context *ctx)
{
	ctx->total[0] = 0;
	ctx->total[1] = 0;

	ctx->state[0] = 0x7380166F;
	ctx->state[1] = 0x4914B2B9;
	ctx->state[2] = 0x172442D7;
	ctx->state[3] = 0xDA8A0600;
	ctx->state[4] = 0xA96F30BC;
	ctx->state[5] = 0x163138AA;
	ctx->state[6] = 0xE38DEE4D;
	ctx->state[7] = 0xB0FB0E4E;
}

static void sm3_process(struct sm3_context *ctx, const u8 data[64])
{
	u32 SS1, SS2, TT1, TT2, W[68], W1[64];
	u32 A, B, C, D, E, F, G, H;
	u32 T[64];
	u32 Temp1, Temp2, Temp3, Temp4, Temp5;
	int j;

	for (j = 0; j < 16; j++)
		T[j] = 0x79CC4519;
	for (j = 16; j < 64; j++)
		T[j] = 0x7A879D8A;

	GET_UINT32_BE(W[0], data,  0);
	GET_UINT32_BE(W[1], data,  4);
	GET_UINT32_BE(W[2], data,  8);
	GET_UINT32_BE(W[3], data, 12);
	GET_UINT32_BE(W[4], data, 16);
	GET_UINT32_BE(W[5], data, 20);
	GET_UINT32_BE(W[6], data, 24);
	GET_UINT32_BE(W[7], data, 28);
	GET_UINT32_BE(W[8], data, 32);
	GET_UINT32_BE(W[9], data, 36);
	GET_UINT32_BE(W[10], data, 40);
	GET_UINT32_BE(W[11], data, 44);
	GET_UINT32_BE(W[12], data, 48);
	GET_UINT32_BE(W[13], data, 52);
	GET_UINT32_BE(W[14], data, 56);
	GET_UINT32_BE(W[15], data, 60);

#define FF0(x, y, z)	((x) ^ (y) ^ (z))
#define FF1(x, y, z)	(((x) & (y)) | ((x) & (z)) | ((y) & (z)))

#define GG0(x, y, z)	((x) ^ (y) ^ (z))
#define GG1(x, y, z)	(((x) & (y)) | ((~(x)) & (z)))

#define SHL(x, n)	((x) << (n))
#define ROTL(x, n)	(SHL((x), (n) & 0x1F) | ((x) >> (32 - ((n) & 0x1F))))

#define P0(x)	((x) ^ ROTL((x), 9) ^ ROTL((x), 17))
#define P1(x)	((x) ^ ROTL((x), 15) ^ ROTL((x), 23))

	for (j = 16; j < 68; j++) {
		/*
		 * W[j] = P1( W[j-16] ^ W[j-9] ^ ROTL(W[j-3],15)) ^
		 *        ROTL(W[j - 13],7 ) ^ W[j-6];
		*/

		Temp1 = W[j - 16] ^ W[j - 9];
		Temp2 = ROTL(W[j - 3], 15);
		Temp3 = Temp1 ^ Temp2;
		Temp4 = P1(Temp3);
		Temp5 =  ROTL(W[j - 13], 7) ^ W[j - 6];
		W[j] = Temp4 ^ Temp5;
	}

	for (j =  0; j < 64; j++)
		W1[j] = W[j] ^ W[j + 4];

	A = ctx->state[0];
	B = ctx->state[1];
	C = ctx->state[2];
	D = ctx->state[3];
	E = ctx->state[4];
	F = ctx->state[5];
	G = ctx->state[6];
	H = ctx->state[7];

	for (j = 0; j < 16; j++) {
		SS1 = ROTL(ROTL(A, 12) + E + ROTL(T[j], j), 7);
		SS2 = SS1 ^ ROTL(A, 12);
		TT1 = FF0(A, B, C) + D + SS2 + W1[j];
		TT2 = GG0(E, F, G) + H + SS1 + W[j];
		D = C;
		C = ROTL(B, 9);
		B = A;
		A = TT1;
		H = G;
		G = ROTL(F, 19);
		F = E;
		E = P0(TT2);
	}

	for (j = 16; j < 64; j++) {
		SS1 = ROTL(ROTL(A, 12) + E + ROTL(T[j], j), 7);
		SS2 = SS1 ^ ROTL(A, 12);
		TT1 = FF1(A, B, C) + D + SS2 + W1[j];
		TT2 = GG1(E, F, G) + H + SS1 + W[j];
		D = C;
		C = ROTL(B, 9);
		B = A;
		A = TT1;
		H = G;
		G = ROTL(F, 19);
		F = E;
		E = P0(TT2);
	}

	ctx->state[0] ^= A;
	ctx->state[1] ^= B;
	ctx->state[2] ^= C;
	ctx->state[3] ^= D;
	ctx->state[4] ^= E;
	ctx->state[5] ^= F;
	ctx->state[6] ^= G;
	ctx->state[7] ^= H;
}

static void sm3_update(struct sm3_context *ctx, const u8 *input, size_t ilen)
{
	size_t fill;
	size_t left;

	if (!ilen)
		return;

	left = ctx->total[0] & 0x3F;
	fill = 64 - left;

	ctx->total[0] += ilen;

	if (unlikely(ctx->total[0] < ilen))
		ctx->total[1]++;

	if (unlikely(left && ilen >= fill)) {
		memcpy(ctx->buffer + left, input, fill);
		sm3_process(ctx, ctx->buffer);
		input += fill;
		ilen -= fill;
		left = 0;
	}

	while (ilen >= 64) {
		sm3_process(ctx, input);
		input += 64;
		ilen -= 64;
	}

	if (ilen > 0)
		memcpy(ctx->buffer + left, input, ilen);
}

static const u8 sm3_padding[64] = {
	0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

static void sm3_final(struct sm3_context *ctx, u8 output[32])
{
	u32 last, padn;
	u32 high, low;
	u8 msglen[8];

	high = (ctx->total[0] >> 29) | (ctx->total[1] <<  3);
	low  = ctx->total[0] << 3;

	PUT_UINT32_BE(high, msglen, 0);
	PUT_UINT32_BE(low,  msglen, 4);

	last = ctx->total[0] & 0x3F;
	padn = (last < 56) ? (56 - last) : (120 - last);

	sm3_update(ctx, sm3_padding, padn);
	sm3_update(ctx, msglen, 8);

	PUT_UINT32_BE(ctx->state[0], output,  0);
	PUT_UINT32_BE(ctx->state[1], output,  4);
	PUT_UINT32_BE(ctx->state[2], output,  8);
	PUT_UINT32_BE(ctx->state[3], output, 12);
	PUT_UINT32_BE(ctx->state[4], output, 16);
	PUT_UINT32_BE(ctx->state[5], output, 20);
	PUT_UINT32_BE(ctx->state[6], output, 24);
	PUT_UINT32_BE(ctx->state[7], output, 28);
}

static volatile int running = 0;
static u8 sm2_rng_pool[64] = {0};
static u32 sm2_rng_index = 0;
static struct sm3_context sm2_rng_ctx;
static struct task_struct *sm2_task = NULL;

static sm2p256v1_params_t sm2p256v1_params = {
	.p = {0xFFFFFFFF,0xFFFFFFFF,0x00000000,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFE},
	.p_h = {0x00000003,0x00000002,0xFFFFFFFF,0x00000002,0x00000001,0x00000001,0x00000002,0x00000004},
	.a = {0xFFFFFFFC,0xFFFFFFFF,0x00000000,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFE},
	.b = {0x4D940E93,0xDDBCBD41,0x15AB8F92,0xF39789F5,0xCF6509A7,0x4D5A9E4B,0x9D9F5E34,0x28E9FA9E},
	.Gx = {0x334C74C7,0x715A4589,0xF2660BE1,0x8FE30BBF,0x6A39C994,0x5F990446,0x1F198119,0x32C4AE2C},
	.Gy = {0x2139F0A0,0x02DF32E5,0xC62A4740,0xD0A9877C,0x6B692153,0x59BDCEE3,0xF4F6779C,0xBC3736A2},
	.n = {0x39D54123,0x53BBF409,0x21C6052B,0x7203DF6B,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFE},
	.n_h = {0x7C114F20,0x901192AF,0xDE6FA2FA,0x3464504A,0x3AFFE0D4,0x620FC84C,0xA22B3D3B,0x1EB5E412},
	.n_1 = {0x39D54122,0x53BBF409,0x21C6052B,0x7203DF6B,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFE},
	.Gx_2_128 = {0xD13A42ED,0xEAE3D9A9,0x484E1B38,0x2B2308F6,0x88C21F3A,0x3DB7B248,0x74D55DA9,0xB692E5B5},
	.Gy_2_128 = {0xE295E5AB,0xD186469D,0x73438E6D,0xDB61AC17,0x544926F9,0x5A924F85,0x0F3FB613,0xA175051B},
};

static inline int get_valid_words(u32 *a, int max_words)
{
	int i;

	for (i = max_words; i > 0; i--) {
		if (a[i - 1])
			return i;
	}

	return 0;
}

static inline int BigNumCmp(u32 *a, int aWordLen, u32 *b, int bWordLen)
{
	int i;

	aWordLen = get_valid_words(a, aWordLen);
	bWordLen = get_valid_words(b, bWordLen);

	if (aWordLen > bWordLen)
		return 1;

	if (aWordLen < bWordLen)
		return -1;

	for (i = (aWordLen-1); i >= 0; i--) {
		if (a[i] > b[i])
			return 1;

		if (a[i] < b[i])
			return -1;
	}

	return 0;
}

static inline int BigNum_Check_Zero_U32(u32 *a, int aWordLen)
{
	int i;

	for (i = 0; i < aWordLen; i++) {
		if (a[i])
			return 0;
	}

	return 1;
}

static inline int BigNum_Check_Zero_U8(void *a, int aByteLen)
{
	int i;
	u64 *b = (u64*)a;
	u8 *c = (u8*)a;
	int dwordlen = aByteLen >> 3;

	for (i = 0; i < dwordlen; i++) {
		if (b[i])
			return 0;
	}

	for (i = (dwordlen << 3); i < aByteLen; i++) {
		if (c[i])
			return 0;
	}

	return 1;
}


static inline void BigNum_Add_One(u32 *a, int wordLen)
{
	int i;

	for (i = 0; i < wordLen; i++) {
		a[i] += 1;
		if (a[i] > 0)
			break;
	}
}

static inline void BigNum_Convert_Byte_Order_U256(const void *in, void *out)
{
	u32 i, tmp;
	const u32 *input = in;
	u32 *output = out;

	for (i = 0; i < 4; i++) {
		tmp = input[i];
		output[i] = swap32(input[0x7 - i]);
		output[0x7 - i] = swap32(tmp);
	}
}


static inline volatile u32 * PKE_A(volatile pke_reg_t *pke_reg, unsigned long a, int step)
{
	volatile u32  *addr;

	addr = (volatile u32 *)((unsigned long)pke_reg + 0x0400 + a * step);

	return addr;	
}

static inline volatile u32 * PKE_B(volatile pke_reg_t *pke_reg, unsigned long a, int step)
{
	volatile u32  *addr;

	addr = (volatile u32 *)((unsigned long)pke_reg + 0x1000 + a * step);

	return addr;	
}

static inline void pke_load_operand(volatile u32 *baseaddr, u32 *data, int wordLen)
{
	int i;

	for (i = 0; i < wordLen; i++) {
		baseaddr[i] = data[i];
	}
}

static inline void pke_read_operand(volatile u32 *baseaddr, u32 *data, int wordLen)  
{
	int i;

	for (i = 0; i < wordLen; i++) {
		data[i] = baseaddr[i];
	}
}

static inline void pke_set_microcode(volatile pke_reg_t *pke_reg, u32 addr)
{
	pke_reg->mc_ptr = addr;
}

static inline void pke_clear_interrupt(volatile pke_reg_t *pke_reg)
{
	pke_reg->stat = 0x01;
}

static inline void pke_start(volatile pke_reg_t *pke_reg)
{
	pke_reg->ctrl = PKE_START_CALC;
}

static inline int pke_check_rt_code(volatile pke_reg_t *pke_reg)
{
	int ret;

	ret = pke_reg->rt_code & 0x07;
	if (unlikely(ret))
		printk(KERN_ERR"rt_code[0x%x] err!\n", ret);

	return ret;
}

static inline void pke_load_pre_calc_mont_H(volatile pke_reg_t *pke_reg, u32 *H,  u32 wordLen)
{
	pke_load_operand(PKE_B(pke_reg, 0, SM2_MEM_STEP), H, wordLen);
}

static u32 trng_check_ready(void)
{
	static u32 count = 0;
	u32 rng_byte = 0;

	if (scto.trng_reg->fifo_sr & 0x1000000) {
		if ((scto.trng_reg->sr & 1) || (count++ > 65536)) {
			scto.trng_reg->cr = 0x1E;

			scto.trng_reg->sr = 7;   //clear flag

			scto.trng_reg->rtcr = 0;

			dsb(sy);

			scto.trng_reg->cr = 0x1F;

			dsb(sy);

			printk(KERN_WARNING"trng timeout!\n");
		}	
	} else {
		rng_byte = (scto.trng_reg->fifo_sr & 0xFC0000) >> 16;
		if (rng_byte)
			count = 0;
	}

	//return counts of random in FIFO
	return rng_byte;
}

static u32 get_rand(u8 *a, u32 byteLen)
{
	u32 rng_data, rng_data_len, rng_byte = 0;

	if (byteLen) {
		rng_data_len = trng_check_ready();
		while (rng_data_len) {
			rng_data = scto.trng_reg->dr;
			if (byteLen >= 4) {
				memcpy(a, &rng_data, 4);
				//printk(KERN_INFO"rng:0x%x\n", rng_data);
				rng_byte += 4;
				a += 4;
				byteLen -= 4;
				rng_data_len -= 4;
			} else {
				memcpy(a, &rng_data, byteLen);
				rng_byte += byteLen;
				byteLen = 0;
				break;
			}
		}
	}

	return rng_byte;
}

static void crypto_rng_update(void)
{
	u32 len, rng_byte;
	u8 *addr;

	addr = sm2_rng_pool + sm2_rng_index;
	len = 64 - sm2_rng_index;

	mutex_lock(&scto.rng_lock);
	rng_byte = get_rand(addr, len);
	mutex_unlock(&scto.rng_lock);

	sm2_rng_index = (sm2_rng_index + rng_byte) & 0x3F;
}

static int sm2_pointMul_Shamir(u32 *k1, u32 *P1x, u32 *P1y,
									u32 *k2, u32 *P2x, u32 *P2y,
									u32 *Qx, u32 *Qy,
									volatile pke_reg_t *pke_reg,
									int work_step)
{
	int i;

	switch (work_step) {
		case 0:
			break;
		case 1:
			goto STEP_1;
		case 2:
			goto STEP_2;
		default:
			printk(KERN_ERR"%s:work_step[0x%x] err!\n", __func__, work_step);
			return -1;
	}

	pke_load_operand(PKE_A(pke_reg, 0, SM2_MEM_STEP), sm2p256v1_params.p, SM2_WORD_LEN);    //p
	pke_set_microcode(pke_reg, MICROCODE_MGMR_PRE_N0);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 1;

STEP_1:
	if ((pke_reg->stat & 1) == 0)
		return 1;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_load_operand(PKE_B(pke_reg, 0, SM2_MEM_STEP), sm2p256v1_params.p_h, SM2_WORD_LEN);  //p_h
	pke_load_operand(PKE_B(pke_reg, 1, SM2_MEM_STEP), P1x, SM2_WORD_LEN);                        //P1x
	pke_load_operand(PKE_B(pke_reg, 2, SM2_MEM_STEP), P1y, SM2_WORD_LEN);                        //P1y                           //P1z=1
	*(PKE_A(pke_reg, 3, SM2_MEM_STEP)) = 1;
	for (i = 1; i < SM2_WORD_LEN; i++)
		*(PKE_A(pke_reg, 3, SM2_MEM_STEP) + i) = 0;
	pke_load_operand(PKE_B(pke_reg, 5, SM2_MEM_STEP), P2x, SM2_WORD_LEN);                        //P2x
	pke_load_operand(PKE_B(pke_reg, 6, SM2_MEM_STEP), P2y, SM2_WORD_LEN);                        //P2y
	pke_load_operand(PKE_B(pke_reg, 4, SM2_MEM_STEP), sm2p256v1_params.a, SM2_WORD_LEN);    //a
	pke_load_operand(PKE_A(pke_reg, 4, SM2_MEM_STEP), k1, SM2_WORD_LEN);                         //k1
	pke_load_operand(PKE_A(pke_reg, 5, SM2_MEM_STEP), k2, SM2_WORD_LEN);                         //k2
	pke_set_microcode(pke_reg, MICROCODE_PMULF);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 2;

STEP_2:
	if ((pke_reg->stat & 1) == 0)
		return 2;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_read_operand(PKE_A(pke_reg, 1, SM2_MEM_STEP), Qx, SM2_WORD_LEN);
	if (Qy != NULL)
		pke_read_operand(PKE_A(pke_reg, 2, SM2_MEM_STEP), Qy, SM2_WORD_LEN);

	return 0;
}

static int sm2_pointMul_base(u32 *k, u32 *x, u32 *y,
								volatile pke_reg_t *pke_reg,
								int work_step)
{
	u32 k1[SM2_WORD_LEN], k2[SM2_WORD_LEN];
	int i;

	//k1: high half part
	for (i = 0; i < SM2_WORD_LEN/2; i++)
		k1[i] = k[i + SM2_WORD_LEN/2];
	for (i = SM2_WORD_LEN/2; i < SM2_WORD_LEN; i++)
		k1[i] = 0;

	//k2: low half part
	for (i = 0; i < SM2_WORD_LEN/2; i++)
		k2[i] = k[i];
	for (i = SM2_WORD_LEN/2; i < SM2_WORD_LEN; i++)
		k2[i] = 0;

	return sm2_pointMul_Shamir(k1, sm2p256v1_params.Gx_2_128, sm2p256v1_params.Gy_2_128,
							k2, sm2p256v1_params.Gx, sm2p256v1_params.Gy,
							x, y, pke_reg, work_step);
}

int sm2_pointMul(u32 *k, u32 *Px, u32 *Py, u32 *Qx, u32 *Qy,
					volatile pke_reg_t *pke_reg, int work_step)
{
	int i;

	switch (work_step) {
		case 0:
			break;
		case 1:
			goto STEP_1;
		case 2:
			goto STEP_2;
		default:
			printk(KERN_ERR"%s:work_step[0x%x] err!\n", __func__, work_step);
			return -1;
	}

	pke_load_operand(PKE_A(pke_reg, 0, SM2_MEM_STEP), sm2p256v1_params.p, SM2_WORD_LEN);    //p
	pke_set_microcode(pke_reg, MICROCODE_MGMR_PRE_N0);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 1;

STEP_1:
	if ((pke_reg->stat & 1) == 0)
		return 1;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_load_operand(PKE_B(pke_reg, 0, SM2_MEM_STEP), sm2p256v1_params.p_h, SM2_WORD_LEN);  //p_h
	pke_load_operand(PKE_B(pke_reg, 1, SM2_MEM_STEP), Px, SM2_WORD_LEN);                         //Px
	pke_load_operand(PKE_B(pke_reg, 2, SM2_MEM_STEP), Py, SM2_WORD_LEN);                         //Py                            //Pz=1
	*((uint32_t *)PKE_A(pke_reg, 3, SM2_MEM_STEP)) = 1;
	for (i = 1; i < SM2_WORD_LEN; i++)
		*(PKE_A(pke_reg, 3, SM2_MEM_STEP) + i) = 0;
	pke_load_operand(PKE_B(pke_reg, 4, SM2_MEM_STEP), sm2p256v1_params.a, SM2_WORD_LEN);    //a
	pke_load_operand(PKE_A(pke_reg, 4, SM2_MEM_STEP), k, SM2_WORD_LEN);                          //k	
	pke_set_microcode(pke_reg, MICROCODE_PMUL);		
	pke_clear_interrupt(pke_reg);		
	pke_start(pke_reg);
	return 2;

STEP_2:
	if ((pke_reg->stat & 1) == 0)
		return 2;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_read_operand(PKE_A(pke_reg, 1, SM2_MEM_STEP), Qx, SM2_WORD_LEN);
	if (Qy != NULL)
		pke_read_operand(PKE_A(pke_reg, 2, SM2_MEM_STEP), Qy, SM2_WORD_LEN);

	return 0;
}

static int sm2_modadd(u32 *modulus, u32 *a, u32 *b, u32 *out, u32 wordLen,
						volatile pke_reg_t *pke_reg, int work_step)
{						
	switch (work_step) {
		case 0:
			break;
		case 1:
			goto STEP_1;
		default:
			printk(KERN_ERR"%s:work_step[0x%x] err!\n", __func__, work_step);
			return -1;
	}

	pke_load_operand(PKE_A(pke_reg, 0, SM2_MEM_STEP), modulus, wordLen); //A0 modulus
	pke_load_operand(PKE_A(pke_reg, 1,  SM2_MEM_STEP), a, wordLen);       //A1 a
	pke_load_operand(PKE_B(pke_reg, 1, SM2_MEM_STEP), b, wordLen);       //B1 b	
	pke_set_microcode(pke_reg, MICROCODE_MODADD);	
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 1;

STEP_1:
	if ((pke_reg->stat & 1) == 0)
		return 1;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_read_operand(PKE_A(pke_reg, 1, SM2_MEM_STEP), out, wordLen);                //A1 result

	return 0;
}

static int sm2_modsub(u32 *modulus, u32 *a, u32 *b, u32 *out, u32 wordLen,
						volatile pke_reg_t *pke_reg, int work_step)
{						
	switch (work_step) {
		case 0:
			break;
		case 1:
			goto STEP_1;
		default:
			printk(KERN_ERR"%s:work_step[0x%x] err!\n", __func__, work_step);
			return -1;
	}

	pke_load_operand(PKE_A(pke_reg, 0, SM2_MEM_STEP), modulus, wordLen); //A0 modulus
	pke_load_operand(PKE_A(pke_reg, 1, SM2_MEM_STEP), a, wordLen);       //A1 a
	pke_load_operand(PKE_B(pke_reg, 1, SM2_MEM_STEP), b, wordLen);       //B1 b
	pke_set_microcode(pke_reg, MICROCODE_MODSUB);	
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 1;

STEP_1:
	if ((pke_reg->stat & 1) == 0)
	return 1;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_read_operand(PKE_A(pke_reg, 1, SM2_MEM_STEP), out, wordLen);                 //A1 result

	return 0;
}

static int sm2_modinv(u32 *modulus, u32 *a, u32 *ainv, int modWordLen, int aWordLen,
						volatile pke_reg_t *pke_reg, int work_step)
{
	int i;

	switch (work_step) {
		case 0:
			break;
		case 1:
			goto STEP_1;
		default:
			printk(KERN_ERR"%s:work_step[0x%x] err!\n", __func__, work_step);
			return -1;
	}


	pke_load_operand(PKE_A(pke_reg, 0, SM2_MEM_STEP), modulus, modWordLen); //A0 modulus
	pke_load_operand(PKE_B(pke_reg, 1, SM2_MEM_STEP), a, aWordLen);         //B1 a
	if (modWordLen > aWordLen) {
		for (i = aWordLen; i < modWordLen; i++)
			*(PKE_B(pke_reg, 1, SM2_MEM_STEP) + i) = 0;
	}
	pke_set_microcode(pke_reg, MICROCODE_MODINV);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 1;

STEP_1:
	if ((pke_reg->stat & 1) == 0)
		return 1;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_read_operand(PKE_A(pke_reg, 1, SM2_MEM_STEP), ainv, modWordLen);               //A1 ainv

	return 0;
}

static int sm2_modmul_internal(u32 *modulus, u32 *a, u32 *b, u32 *out, int wordLen,
									volatile pke_reg_t *pke_reg, int work_step)
{
	switch (work_step) {
		case 0:
			break;
		case 1:
			goto STEP_1;
		case 2:
			goto STEP_2;
		default:
			printk(KERN_ERR"%s:work_step[0x%x] err!\n", __func__, work_step);
			return -1;
	}

	pke_load_pre_calc_mont_H(pke_reg, sm2p256v1_params.n_h, SM2_WORD_LEN);
	pke_load_operand(PKE_A(pke_reg, 0, SM2_MEM_STEP), modulus, wordLen); //A0 modulus
	pke_set_microcode(pke_reg, MICROCODE_MGMR_PRE_N0);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 1;

STEP_1:
	if ((pke_reg->stat & 1) == 0)
		return 1;

	if (pke_check_rt_code(pke_reg))
		return -1;


	pke_load_operand(PKE_A(pke_reg, 1,  SM2_MEM_STEP), a, wordLen);       //A1 a
	pke_load_operand(PKE_B(pke_reg, 1, SM2_MEM_STEP), b, wordLen);       //B1 b	
	pke_set_microcode(pke_reg, MICROCODE_MODMUL);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 2;

STEP_2:
	if ((pke_reg->stat & 1) == 0)
		return 2;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_read_operand(PKE_A(pke_reg, 1, SM2_MEM_STEP), out, wordLen);                 //A0 out

	return 0;
}

int sm2_pointVerify(u32 *Px, u32 *Py, volatile pke_reg_t *pke_reg, int work_step)
{
	switch (work_step) {
		case 0:
			break;
		case 1:
			goto STEP_1;
		case 2:
			goto STEP_2;
		default:
			printk(KERN_ERR"%s:work_step[0x%x] err!\n", __func__, work_step);
			return -1;
	}

	pke_load_operand(PKE_A(pke_reg, 0, SM2_MEM_STEP), sm2p256v1_params.p, SM2_WORD_LEN);   //p
	pke_set_microcode(pke_reg, MICROCODE_MGMR_PRE_N0);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 1;

STEP_1:
	if ((pke_reg->stat & 1) == 0)
		return 1;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_load_operand(PKE_B(pke_reg, 0, SM2_MEM_STEP), sm2p256v1_params.p_h, SM2_WORD_LEN); //p_h
	pke_load_operand(PKE_A(pke_reg, 1, SM2_MEM_STEP), Px, SM2_WORD_LEN);            //Px
	pke_load_operand(PKE_A(pke_reg, 2, SM2_MEM_STEP), Py, SM2_WORD_LEN);            //Py
	pke_load_operand(PKE_B(pke_reg, 4, SM2_MEM_STEP), sm2p256v1_params.a, SM2_WORD_LEN);   //a
	pke_load_operand(PKE_A(pke_reg, 4, SM2_MEM_STEP), sm2p256v1_params.b, SM2_WORD_LEN);   //b
	pke_set_microcode(pke_reg, MICROCODE_PVER);		
	pke_clear_interrupt(pke_reg);	
	pke_start(pke_reg);
	return 2;

STEP_2:
	if ((pke_reg->stat & 1) == 0)
		return 2;

	if (pke_check_rt_code(pke_reg))
		return -1;

	return 0;
}


static inline void sm2_kdf_counter_add_one(u8 a[4])
{
	int i;

	for (i = 3; i >= 0; i--) {
		a[i] += 1;
		if (a[i] > 0)
			break;
	}
}

static void sm2_kdf(u8 *x2 , u8 *y2, u8 *k, int kLen)
{
	u8 digest[SM3_DIGEST_BYTE_LEN];
	int i, t;
	u8 counter[4] = {0x00,0x00,0x00,0x01};      // count = 1;
	struct sm3_context basectx, ctx;

	sm3_init(&basectx);		
	sm3_update(&basectx, x2, SM2_BYTE_LEN);
	sm3_update(&basectx, y2, SM2_BYTE_LEN);

	t = kLen >> 5;                                 // t = kByteLen/32;
	for (i = 0; i < t; i++) {
		memcpy(&ctx, &basectx, sizeof(ctx));
		sm3_update(&ctx, counter, 4);
		sm3_final(&ctx, k + (i << 5));
		sm2_kdf_counter_add_one(counter);
	}

	kLen = kLen & 0x1F;
	if (kLen) {
		memcpy(&ctx, &basectx, sizeof(ctx));
		sm3_update(&ctx, counter, 4);
		sm3_final(&ctx, digest);
		memcpy(k + (t << 5), digest, kLen);
	}
}

static inline void XOR(const void *A, const void *B, void *C, int byteLen)
{
	int i;
	int dwordlen = byteLen >> 3;
	u64 *dA = (u64 *)A, *dB = (u64 *)B, *dC = (u64 *)C;
	u8 *bA = (u8 *)A, *bB = (u8 *)B, *bC = (u8 *)C;

	for (i = 0; i < dwordlen; i++)
		dC[i] = dA[i] ^ dB[i];

	for (i = (dwordlen << 3); i < byteLen; i++)
		bC[i] = bA[i] ^ bB[i];
}

static void sm2_rng(u32 k[8])
{
	crypto_rng_update();
READ_RNG_K:
	/*
	  用随机数发生器产生随机数k ∈[1,n-1]；
	*/
	sm3_update(&sm2_rng_ctx, sm2_rng_pool, 64);
	memcpy(k, sm2_rng_ctx.state, SM2_BYTE_LEN);
	memcpy(sm2_rng_pool, k, SM2_BYTE_LEN);

	//printk(KERN_INFO"k:0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x\n", k[0], k[1], k[2], k[3], k[4], k[5], k[6], k[7]);

	if (BigNum_Check_Zero_U32(k, SM2_WORD_LEN))
		goto READ_RNG_K;
	if (BigNumCmp(k, SM2_WORD_LEN, sm2p256v1_params.n, SM2_WORD_LEN) >= 0)
		goto READ_RNG_K;
}

static int __sm2_sign(u8 *r_s, volatile pke_reg_t *pke_reg, phytium_sm2_work_list_t *work_list)
{
	int ret;

	switch (work_list->work_step >> 8) {
		case 0:
			break;
		case 1:
			goto STEP_1;
		case 2:
			goto STEP_2;
		case 3:
			goto STEP_3;
		case 4:
			goto STEP_4;
		case 5:
			goto STEP_5;
		case 6:
			goto STEP_6;
		case 7:
			goto STEP_7;
		default:
			printk(KERN_ERR"%s:work_step[0x%x] err!\n", __func__, work_list->work_step);
			return -1;
	}
	//e
	/*
	 5.5 用户其它信息
	     作为签名者的用户A具有长度为entlenA比特的可辨别标识IDA，记ENTLA是由整数entlenA转换
	 而成的两个字节，在本部分规定的椭圆曲线数字签名算法中，签名者和验证者都需要用密码杂凑函
	 数求得用户A的杂凑值ZA。按本文本第1部分4.2.5和4.2.4给出的细节，将椭圆曲线方程参数a、 b、 G的
	 坐标xG、 yG 和PA的坐标xA、 yA的数据类型转换为比特串， ZA=H256(ENTLA || IDA || a || b || xG ||
	 yG || xA || yA)。
	 6.1 数字签名的生成算法
	     A1：置M=ZA|| M；
	     A2：计算e = H(M)，按本文本第1部分4.2.3和4.2.2给出的细节将e的数据类型转换为整数；
	*/
	//dA
	//make sure dA in [1, n-2]
	/*
	 6.1 密钥对的生成
	     输入：一个有效的Fq(q = p且p为大于3的素数，或q = 2m)上椭圆曲线系统参数的集合。
	     输出：与椭圆曲线系统参数相关的一个密钥对(d;P)。
	           a) 用随机数发生器产生整数d ∈ [1;n − 2]；
	           b) G为基点，计算点P = (xP;yP) = [d]G； (参见附录A.3.2)
	           c) 密钥对是(d;P)，其中d为私钥， P为公钥。
	     这里不做检查，默认私钥输入正确。
	*/

READ_RNG_K:
	/*
	 A3：用随机数发生器产生随机数k ∈[1,n-1]；
	*/
	sm2_rng(work_list->k);
	/*
	 A4：计算椭圆曲线点(x1,y1)=[k]G；
	*/

STEP_1:
	ret = sm2_pointMul_base(work_list->k,
							work_list->x1,
							NULL,
							pke_reg,
							work_list->work_step & 0xFF);

	work_list->work_step = ret | (1 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:sm2_pointMul_base fail!ret:%d\n", __func__, ret);
		return ret;
	}

	/*
	 A5：计算r=(e+x1) modn，若r=0或r+k=n则返回A3；
	*/
STEP_2:
	ret = sm2_modadd(sm2p256v1_params.n,
					work_list->e,
					work_list->x1,
					work_list->r,
					SM2_WORD_LEN,
					pke_reg,
					work_list->work_step & 0xFF);

	work_list->work_step = ret | (2 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:sm2_modadd fail!ret:%d\n", __func__, ret);
		return ret;
	}

	//make sure r is not zero
	if (BigNum_Check_Zero_U32(work_list->r, SM2_WORD_LEN))
		goto READ_RNG_K;

	//tmp1 = r + k mod n
STEP_3:
	ret = sm2_modadd(sm2p256v1_params.n,
					work_list->r,
					work_list->k,
					work_list->tmp1,
					SM2_WORD_LEN,
					pke_reg,
					work_list->work_step & 0xFF);

	work_list->work_step = ret | (3 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:sm2_modadd fail!ret:%d\n", __func__, ret);
		return ret;
	}

	//make sure r+k is not n
	if (BigNum_Check_Zero_U32(work_list->tmp1, SM2_WORD_LEN))
		goto READ_RNG_K;

	/*
	A6：计算s = ((1 + dA)^(−1) · (k − r · dA)) modn，若s=0则返回A3；
	*/
	//tmp1 =  r*dA mod n
STEP_4:
	ret = sm2_modmul_internal(sm2p256v1_params.n,
							work_list->r,
							work_list->ctx->d,
							work_list->tmp1,
							SM2_WORD_LEN,
							pke_reg,
							work_list->work_step & 0xFF);

	work_list->work_step = ret | (4 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:sm2_modmul_internal fail!ret:%d\n", __func__, ret);
		return ret;
	}

	//tmp1 =  (k - r*dA) mod n
STEP_5:
	ret = sm2_modsub(sm2p256v1_params.n,
					work_list->k,
					work_list->tmp1,
					work_list->tmp1,
					SM2_WORD_LEN,
					pke_reg,
					work_list->work_step & 0xFF);

	work_list->work_step = ret | (5 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:sm2_modsub fail!ret:%d\n", __func__, ret);
		return ret;
	}

	//tmp2 = (1+dA)^(-1) mod n
	memcpy(work_list->tmp2, work_list->ctx->d, SM2_BYTE_LEN);
	BigNum_Add_One(work_list->tmp2, SM2_WORD_LEN);

STEP_6:
	ret = sm2_modinv(sm2p256v1_params.n,
					work_list->tmp2,
					work_list->tmp2,
					SM2_WORD_LEN,
					SM2_WORD_LEN,
					pke_reg,
					work_list->work_step & 0xFF);

	work_list->work_step = ret | (6 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:sm2_modinv fail!ret:%d\n", __func__, ret);
		return ret;
	}

	//s = ((1+dA)^(-1))*(k - r*dA) mod n
STEP_7:
	ret = sm2_modmul_internal(sm2p256v1_params.n,
							work_list->tmp1,
							work_list->tmp2,
							work_list->s,
							SM2_WORD_LEN,
							pke_reg,
							work_list->work_step & 0xFF);

	work_list->work_step = ret | (7 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:sm2_modmul_internal fail!ret:%d\n", __func__, ret);
		return ret;
	}

	//make sure s is not zero
	if (BigNum_Check_Zero_U32(work_list->s, SM2_WORD_LEN))
		goto READ_RNG_K;

	BigNum_Convert_Byte_Order_U256(work_list->r, r_s);
	BigNum_Convert_Byte_Order_U256(work_list->s, r_s + SM2_BYTE_LEN);

	return 0;
}

static int __sm2_verify(volatile pke_reg_t *pke_reg, phytium_sm2_work_list_t *work_list)
{
	int ret;

	switch (work_list->work_step >> 8) {
		case 0:
			break;
		case 1:
			goto STEP_1;
		case 2:
			goto STEP_2;
		case 3:
			goto STEP_3;
		default:
			printk(KERN_ERR"%s:work_step[0x%x] err!\n", __func__, work_list->work_step);
			return -1;
	}

	/*
	 B5：计算t = (r + s) modn，若t = 0，则验证不通过；
	*/
STEP_1:
	ret = sm2_modadd(sm2p256v1_params.n,
					work_list->r,
					work_list->s,
					work_list->t,
					SM2_WORD_LEN,
					pke_reg,
					work_list->work_step & 0xFF);

	work_list->work_step = ret | (1 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:sm2_modadd fail!ret:%d\n", __func__, ret);
		return ret;
	}
	if (BigNum_Check_Zero_U32(work_list->t, SM2_WORD_LEN)){
		printk(KERN_ERR"%s:t == zero, err!\n", __func__);
		return -1;
	}

	/*
	 B6：计算椭圆曲线点(x1,y1)=[s]G + [t]PA；
	*/
STEP_2:
	ret = sm2_pointMul_Shamir(work_list->s,
							sm2p256v1_params.Gx,
							sm2p256v1_params.Gy,
							work_list->t,
							work_list->ctx->x,
							work_list->ctx->y,
							work_list->x1,
							NULL,
							pke_reg,
							work_list->work_step & 0xFF);

	work_list->work_step = ret | (2 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:sm2_pointMul_Shamir fail!ret:%d\n", __func__, ret);
		return ret;
	}

	/*
	 B3：置M=ZA||M；
	 B4：计算e = H(M)
	*/
	//e

	/*
	 计算R = (e + x1) modn，检验R=r是否成立，若成立则验证通过；否则验证不通过
	*/
STEP_3:
	ret = sm2_modadd(sm2p256v1_params.n,
					work_list->e,
					work_list->x1,
					work_list->R,
					SM2_WORD_LEN,
					pke_reg,
					work_list->work_step & 0xFF);

	work_list->work_step = ret | (3 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:sm2_modadd fail!ret:%d\n", __func__, ret);
		return ret;
	}

	//cmp R == r
	if (BigNumCmp(work_list->R, SM2_WORD_LEN, work_list->r, SM2_WORD_LEN)){
		printk(KERN_ERR"%s:R != r, err!\n", __func__);
		return -1;
	}

	return 0;
}

static int __sm2_encrypt(u8 *M, u32 MLen,
				u8 *out, u32 *outlen,
				volatile pke_reg_t *pke_reg,
				phytium_sm2_work_list_t *work_list)
{
	u8 *C1 = out, *C2 = out + 65, *C3 = out + 65 + MLen;
	int ret;
	struct sm3_context ctx;

	switch (work_list->work_step >> 8) {
		case 0:
			break;
		case 1:
			goto STEP_1;
		case 2:
			goto STEP_2;
		default:
			printk(KERN_ERR"%s:work_step[0x%x] err!\n", __func__, work_list->work_step);
			return -1;
	}

READ_RNG_K:
	/*
	 A1：用随机数发生器产生随机数k ∈[1,n-1]；
	*/
	sm2_rng(work_list->k);

	/*
	 A2: 计算椭圆曲线点C1=[k]G=(x1,y1)
	*/
STEP_1:
	ret = sm2_pointMul_base(work_list->k,
							work_list->x1,
							work_list->y1,
							pke_reg,
							work_list->work_step & 0xFF);

	work_list->work_step = ret | (1 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:sm2_pointMul_base fail!ret:%d\n", __func__, ret);
		return ret;
	}

	C1[0] = POINT_NOT_COMPRESSED;
	BigNum_Convert_Byte_Order_U256(work_list->x1, &C1[1]);
	BigNum_Convert_Byte_Order_U256(work_list->y1, &C1[1 + SM2_BYTE_LEN]);

	/*
	 A3：计算椭圆曲线点S=[h]PB，若S是无穷远点，则报错并退出.
	 A4：计算椭圆曲线点[k]PB=(x2,y2)
	*/
STEP_2:
	ret = sm2_pointMul(work_list->k,
					work_list->ctx->x,
					work_list->ctx->y,
					work_list->x1,
					work_list->y1,
					pke_reg,
					work_list->work_step & 0xFF);

	work_list->work_step = ret | (2 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:sm2_pointMul fail!ret:%d\n", __func__, ret);
		return ret;
	}

	/*
	 A5：计算t=KDF(x2 || y2, klen)，若t为全0比特串，则返回A1；
	*/
	BigNum_Convert_Byte_Order_U256(work_list->x1, work_list->x1);
	BigNum_Convert_Byte_Order_U256(work_list->y1, work_list->y1);
	sm2_kdf((u8 *)work_list->x1, (u8 *)work_list->y1, C2, MLen);
	if (BigNum_Check_Zero_U8(C2, MLen))
		goto READ_RNG_K;

	/*
	 A6：计算C2 = M ⊕ t；
	*/
	XOR(C2, M, C2, MLen);

	/*
	 A7：计算C3 = Hash(x2 || M || y2)
	*/
	sm3_init(&ctx);		
	sm3_update(&ctx, (u8*)work_list->x1, SM2_BYTE_LEN);
	sm3_update(&ctx, M, MLen);
	sm3_update(&ctx, (u8*)work_list->y1, SM2_BYTE_LEN);
	sm3_final(&ctx, C3);

	*outlen = MLen + 1 + 3 * SM2_BYTE_LEN;

	return 0;
}

static int __sm2_decrypt(const u8 *C, int CLen,
				u8 *M, u32 *MLen,
				volatile pke_reg_t *pke_reg,
				phytium_sm2_work_list_t *work_list)
{
	int temLen;
	u8 digest[SM2_BYTE_LEN];
	const u8 *C2, *C3;
	int ret;
	struct sm3_context ctx;

	switch (work_list->work_step >> 8) {
		case 0:
			break;
		case 1:
			goto STEP_1;
		case 2:
			goto STEP_2;
		default:
			printk(KERN_ERR"%s:work_step[0x%x] err!\n", __func__, work_list->work_step);
			return -1;
	}

	/*
	 B1：从C中取出比特串C1，验证C1是否满足椭圆曲线方程，若不满足则报错并退出；
	 B2：计算椭圆曲线点S=[h]C1，若S是无穷远点，则报错并退出；
	*/
	BigNum_Convert_Byte_Order_U256(C + 1, work_list->x1);
	BigNum_Convert_Byte_Order_U256(C + 1 + SM2_BYTE_LEN, work_list->y1);
STEP_1:
	ret = sm2_pointVerify(work_list->x1,
						work_list->y1,
						pke_reg,
						work_list->work_step & 0xFF);

	work_list->work_step = ret | (1 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:sm2_pointVerify fail!ret:%d\n", __func__, ret);
		return ret;
	}

	/*
	 B3：计算[dB]C1=(x2,y2)
	*/
STEP_2:
	ret = sm2_pointMul(work_list->ctx->d,
					work_list->x1,
					work_list->y1,
					work_list->x1,
					work_list->y1,
					pke_reg,
					work_list->work_step & 0xFF);

	work_list->work_step = ret | (2 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:sm2_pointMul fail!ret:%d\n", __func__, ret);
		return ret;
	}
	BigNum_Convert_Byte_Order_U256(work_list->x1, work_list->x1);
	BigNum_Convert_Byte_Order_U256(work_list->y1, work_list->y1);

	/*
	 B4：计算t=KDF(x2 || y2, klen)
	*/
	temLen = CLen - (1 + 3 * SM2_BYTE_LEN);
	sm2_kdf((u8*)work_list->x1, (u8*)work_list->y1, M, temLen);
	if (BigNum_Check_Zero_U8(M, temLen)){
		printk(KERN_ERR"%s:t == zero, err!\n", __func__);
		return -1;
	}

	/*
	 B5：从C中取出比特串C2，计算M′ = C2 ⊕ t；
	*/
	C2 = C + (1 + 2 * SM2_BYTE_LEN);
	XOR(M, C2, M, temLen);

	/*
	B6：计算u = Hash(x2 || M′ || y2)，从C中取出比特串C3，若u ̸= C3，则报错并退出；
	*/
	C3 = C2 + temLen;
	sm3_init(&ctx);
	sm3_update(&ctx, (u8 *)work_list->x1, SM2_BYTE_LEN);
	sm3_update(&ctx, M, temLen);
	sm3_update(&ctx, (u8 *)work_list->y1, SM2_BYTE_LEN);
	sm3_final(&ctx, digest);
	if (memcmp(C3, digest, SM2_BYTE_LEN))
	 return -1;

	*MLen = temLen;

	return 0;
}

static void sm2signature_encode(u8 *input, u8 *sig, u32 *sig_len)
{
	u32 offset;
	u32 rlen, slen;
	u8 r[32], s[32];

	for (offset = 0; offset < 31; offset++) {
		if (input[offset])
			break;
	}

	rlen = 32 - offset;
	memcpy(r, input + 32 - rlen, rlen);

	for (offset = 32; offset < 63; offset++) {
		if (input[offset])
			break;
	}

	slen = 64 - offset;	
	memcpy(s, input + 64 - slen, slen);

	*sig_len = 6 + rlen + slen;
	sig[0] = 0x30;//Tag 0, ASN1_SEQ(bit 0-4), Value = TLV + TLV + ...(bit5)
	sig[1] = 4 + rlen + slen;//Len 0,
	sig[2] = 0x02;//Tag 1, ASN1_INT(bit 0-4)
	sig[3] = rlen;//Len 1,

	offset = 4;
	if (r[0] & 0x80) {//Highest bit, r is unsigned Integer
		*sig_len += 1;
		sig[1] ++;
		sig[3] ++;
		sig[4] = 0x00;
		offset ++;
	}

	memcpy(sig + offset, r, rlen);
	offset += rlen;

	sig[offset ++] = 0x02;//Tag 2, ASN1_INT(bit 0-4)
	sig[offset ++] = slen;//Len 2,

	if (s[0] & 0x80) {//Highest bit, s is unsigned Integer
		*sig_len += 1;
		sig[1] ++;
		sig[offset - 1] ++;
		sig[offset ++] = 0x00;
	}

	memcpy(sig + offset, s, slen);	
}

static int sm2signature_decode(u8 *sig, u32 sig_len, u8 *out)
{
	u32 offset = 4, len;
	u32 rlen, slen;
	u8 r[32] = {0}, s[32] = {0};

	if (unlikely((sig[0] != 0x30) || (sig[2] != 0x02))) {//check Tag 0 and Tag 1
		printk(KERN_ERR"%s:((sig[0] != 0x30) || (sig[2] != 0x02)), err!\n", __func__);
		return -1;
	}

	len = sig[1] + 2;
	if (unlikely(len != sig_len)) {//check Len 0
		printk(KERN_ERR"%s:len != sig_len, err!\n", __func__);
		return -1;
	}

	rlen = sig[3];//check Len 1
	if (rlen == 0x21) {
		offset ++;
		rlen = 0x20;
	} else if (rlen > 0x21) {
		printk(KERN_ERR"%s:rlen > 0x21, err!\n", __func__);
		return -1;
	}

	memcpy(r + 32 - rlen, sig + offset, rlen);
	offset += rlen;

	if (unlikely(sig[offset++] != 0x02)) {//check Tag 2
		printk(KERN_ERR"%s:Tag 2 != 0x02, err!\n", __func__);
		return -1;
	}

	slen = sig[offset];//check Len 2
	if (slen == 0x21) {
		offset ++;
		slen = 0x20;
	} else if (slen > 0x21) {
		printk(KERN_ERR"%s:slen > 0x21, err!\n", __func__);
		return -1;
	}
	offset ++;

	if (offset + slen == sig_len)//check Value Length
		memcpy(s + 32 - slen, sig + offset, slen);
	else {
		printk(KERN_ERR"%s:offset + slen != sig_len, err!\n", __func__);
		return -1;
	}

	memcpy(out, r, 32);
	memcpy(out + 32, s, 32);

	return 0;
}

static void sm2encrypt_encode(u8 *input, u32 inlen, u8 *output, u32 *outlen)
{
	u8 *c2;
	u8 x[32], y[32], hash[32];
	u8 xhead[4], yhead[4], hashhead[4], c2head[8], head[8];
	u32 offset, length, c2headlen, xheadlen, yheadlen, hashheadlen, headlen, c2len, i;
	u32 xlen, ylen;

	c2 = input + 65;
	memcpy(x, input + 1, 32);
	memcpy(y, input + 33, 32);
	memcpy(hash, input + inlen - 32, 32);

	c2len = inlen - 97;
	c2headlen = 2;
	if (c2len > 127)
		c2headlen = 3;
	if (c2len >255)
		c2headlen = 4;
	if (c2len > 65535)
		c2headlen = 5;
	if (c2len > 16777215)
		c2headlen = 6;

	c2head[0] = 0x04;// Tag, ASN1_OTS
	switch (c2headlen) {
		case 2:
			c2head[1] = c2len;//Len(bit7 = 0)
			break;
		case 3:
			c2head[1] = 0x81;//Len(bit7 = 1, bit0-6 = Len length)
			c2head[2] = c2len;
			break;
		case 4:
			c2head[1] = 0x82;
			c2head[2] = c2len >> 8;
			c2head[3] = c2len;
			break;
		case 5:
			c2head[1] = 0x83;
			c2head[2] = c2len >> 16;
			c2head[3] = c2len >> 8;
			c2head[4] = c2len;
			break;
		case 6:
			c2head[1] = 0x84;
			c2head[2] = c2len >> 24;
			c2head[3] = c2len >> 16;
			c2head[4] = c2len >> 8;
			c2head[5] = c2len;
			break;
		default:
			break;
	}

	for (i = 0; i <31; i++) {
		if(x[i])
			break;
	}
	xlen = 32 - i;
	xheadlen = 2;
	if (x[i] & 0x80)//Highest bit, x is unsigned Integer
		xheadlen = 3;
	xhead[0] = 0x2;//Tag, ASN1_INT
	switch (xheadlen) {
		case 2:
			xhead[1] = xlen;
			break;
		case 3:
			xhead[1] = xlen + 1;
			xhead[2] = 0x00;
			break;
		default:
			break;
	}

	for (i = 0; i <31; i++) {
		if (y[i])
			break;
	}
	ylen = 32 - i;
	yheadlen = 2;
	if (y[i] & 0x80)//Highest bit, y is unsigned Integer
		yheadlen = 3;
	yhead[0] = 0x2;//Tag, ASN1_INT
	switch (yheadlen) {
		case 2:
			yhead[1] = ylen;
			break;
		case 3:
			yhead[1] = ylen + 1;
			yhead[2] = 0x00;
			break;
		default:
			break;
	}

	hashheadlen = 2;
	hashhead[0] = 0x04;// Tag, ASN1_OTS
	hashhead[1] = 0x20;

	length = c2headlen + c2len;
	length += xheadlen + xlen;
	length += yheadlen + ylen;
	length += hashheadlen + 32;

	headlen = 2;
	if (length > 127)
		headlen = 3;
	if (length >255)
		headlen = 4;
	if (length > 65535)
		headlen = 5;
	if (length > 16777215)
		headlen = 6;

	head[0] = 0x30;//Tag, ASN1_SEQ, Value = TLV + TLV + ...
	switch (headlen) {
		case 2:
			head[1] = length;
			break;
		case 3:
			head[1] = 0x81;
			head[2] = length;
			break;
		case 4:
			head[1] = 0x82;
			head[2] = length >> 8;
			head[3] = length;
			break;
		case 5:
			head[1] = 0x83;
			head[2] = length >> 16;
			head[3] = length >> 8;
			head[4] = length;
			break;
		case 6:
			head[1] = 0x84;
			head[2] = length >> 24;
			head[3] = length >> 16;
			head[4] = length >> 8;
			head[5] = length;
			break;
		default:
			break;
	}

	offset = 0;
	memcpy(output + offset, head, headlen);
	offset += headlen;
	memcpy(output + offset, xhead, xheadlen);
	offset += xheadlen;
	memcpy(output + offset, x + 32 - xlen, xlen);
	offset += xlen;
	memcpy(output + offset, yhead, yheadlen);
	offset += yheadlen;
	memcpy(output + offset, y + 32 - ylen, ylen);
	offset += ylen;
	memcpy(output + offset, hashhead, hashheadlen);
	offset += hashheadlen;
	memcpy(output + offset, hash, 32);
	offset += 32;
	memcpy(output + offset, c2head, c2headlen);
	offset += c2headlen;
	memmove(output + offset, c2, c2len);
	offset += c2len;

	*outlen = offset;
}

static int sm2decrypt_decode(u8 *input, u32 inlen, u8 *output, u32 *outlen)
{
	u8 x[32] = {0}, y[32] = {0}, hash[32];
	u32 offset, length, c2headlen, xheadlen, yheadlen, hashheadlen, headlen, c2len, i;
	u32 xlen, ylen;

	if (unlikely(inlen < 45)) {//head:10 hash:32 x:1 y:1 c2:1
		printk(KERN_ERR"%s:inlen[%u], err!\n", __func__, inlen);
		return -1;
	}

	if (unlikely(input[0] != 0x30)) {//check Tag
		printk(KERN_ERR"%s:input[0] != 0x30, err!\n", __func__);
		return -1;
	}
	if (input[1] & 0x80) {
		headlen = 2 + (input[1] & 0x7f);
		if (unlikely(headlen > 6)) {
			printk(KERN_ERR"%s:headlen > 6, err!\n", __func__);
			return -1;
		}

		length = 0;
		for (i = 2; i < headlen; i++) {
			length <<= 8;
			length += input[i]; 
		}
	} else {
		headlen = 2;
		length = input[1];
	}

	if (unlikely(headlen + length != inlen)) {
		printk(KERN_ERR"%s:(headlen + length != inlen), err!\n", __func__);
		return -1;
	}

	offset = headlen;
	if (unlikely(input[offset] != 0x02)) {
		printk(KERN_ERR"%s:(input[offset] != 0x02), err!\n", __func__);
		return -1;
	}

	xlen = input[offset + 1];
	xheadlen = 2;
	if (xlen == 0x21) {
		xheadlen = 3;
		xlen = 0x20;
	} else if (xlen > 0x21) {
		printk(KERN_ERR"%s:(xlen > 0x21), err!\n", __func__);
		return -1;
	}

	offset += xheadlen;
	memcpy(x + 32 - xlen, input + offset, xlen);
	offset += xlen;

	if (unlikely(input[offset] != 0x02)) {
		printk(KERN_ERR"%s:(input[offset] != 0x02), err!\n", __func__);
		return -1;
	}
	ylen = input[offset + 1];
	yheadlen = 2;
	if (ylen == 0x21) {
		yheadlen = 3;
		ylen = 0x20;
	}

	if (unlikely(offset + yheadlen + ylen + 2 > inlen)) {
		printk(KERN_ERR"%s:(offset + yheadlen + ylen + 2 > inlen), err!\n", __func__);
		return -1;
	}

	offset += yheadlen;
	memcpy(y + 32 - ylen, input + offset, ylen);
	offset += ylen;

	if (unlikely((input[offset] != 0x04) || (input[offset + 1] != 0x20))) {
		printk(KERN_ERR"%s:((input[offset] != 0x04) || (input[offset + 1] != 0x20)), err!\n", __func__);
		return -1;
	}
	hashheadlen = 2;
	offset += hashheadlen;

	if (unlikely(offset + 35 > inlen)) {
		printk(KERN_ERR"%s:(offset + 35 > inlen), err!\n", __func__);
		return -1;
	}

	memcpy(hash, input + offset, 32);
	offset += 32;

	if (unlikely(input[offset] != 0x04)) {
		printk(KERN_ERR"%s:(input[offset] != 0x04), err!\n", __func__);
		return -1;
	}
	if (input[offset + 1] & 0x80) {
		c2headlen = 2 + (input[offset + 1] & 0x7f);
		if (unlikely(c2headlen > 6)) {
			printk(KERN_ERR"%s:(c2headlen > 6), err!\n", __func__);
			return -1;
		}

		if (unlikely(offset + c2headlen > inlen)) {
			printk(KERN_ERR"%s:(offset + c2headlen > inlen), err!\n", __func__);
			return -1;
		}

		c2len = 0;
		for (i = 2; i < c2headlen; i++) {
			c2len <<= 8;
			c2len += input[offset + i]; 
		}
	} else {
		c2headlen = 2;
		c2len = input[offset + 1];
	}

	offset += c2headlen;
	if (unlikely(offset + c2len != inlen)) {
		printk(KERN_ERR"%s:(offset + c2len != inlen), err!\n", __func__);
		return -1;
	}

	output[0] = 0x04;
	memcpy(output + 1, x, 32);
	memcpy(output + 33, y, 32);
	memmove(output + 65, input + offset, c2len);
	memcpy(output + 65 + c2len, hash, 32);
	*outlen = c2len + 97;

	return 0;
}

static int sm2_sign(volatile pke_reg_t *pke_reg, phytium_sm2_work_list_t *work_list)
{
	struct akcipher_request *req = work_list->req;
	u8 r_s[64], dercode[72];
	u32 sign_len;
	int ret;

	if (unlikely(work_list->work_step == 0)) {
		/*
		 e check and convert
		*/
		if (unlikely(req->src_len != 32)) {
			printk(KERN_ERR"%s:src_len[0x%x], err!\n", __func__, req->src_len);
			return -1;
		}
		if (unlikely(req->dst_len < 72)) {
			printk(KERN_ERR"%s:dst_len[0x%x], err!\n", __func__, req->dst_len);
			return -1;
		}
		sg_copy_buffer(req->src, sg_nents(req->src), r_s, 32, 0, true);
		BigNum_Convert_Byte_Order_U256(r_s, work_list->e);
	}

	ret = __sm2_sign(r_s, pke_reg, work_list);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:__sm2_sign failed!ret:%d\n", __func__, ret);
		return ret;
	}

	sm2signature_encode(r_s, dercode, &sign_len);
	sg_copy_buffer(req->dst, sg_nents(req->dst), dercode, sign_len, 0, false);
	req->dst_len = sign_len;

	return ret;
}

static int sm2_verify(volatile pke_reg_t *pke_reg, phytium_sm2_work_list_t *work_list)
{
	struct akcipher_request *req = work_list->req;
	u8 r_s[64], code[128];
	int ret;

	if (unlikely(work_list->work_step == 0)) {
		/*
		 e check and convert
		*/
		if (unlikely(req->src_len > 72)) {
			printk(KERN_ERR"%s:src_len[0x%x], err!\n", __func__, req->src_len);
			return -1;
		}
		if (unlikely(req->dst_len != 32)) {
			printk(KERN_ERR"%s:dst_len[0x%x], err!\n", __func__, req->dst_len);
			return -1;
		}
		sg_copy_buffer(req->src, sg_nents(req->src), code, req->src_len + req->dst_len, 0, true);
		BigNum_Convert_Byte_Order_U256(code + req->src_len, work_list->e);
		ret = sm2signature_decode(code, req->src_len, r_s);
		if (unlikely(ret)) {
			printk(KERN_ERR"%s:sm2signature_decode failed!ret:%d\n", __func__, ret);
			return ret;
		}
		/*
		 B1：检验r ∈[1,n-1]是否成立，若不成立则验证不通过；
		*/
		BigNum_Convert_Byte_Order_U256(r_s, work_list->r);
		if (unlikely(BigNum_Check_Zero_U32(work_list->r, SM2_WORD_LEN))) {
			printk(KERN_ERR"%s:r == 0, err!\n", __func__);
			return -1;
		}
		if (unlikely(BigNumCmp(work_list->r, SM2_WORD_LEN, sm2p256v1_params.n, SM2_WORD_LEN) >= 0)) {
			printk(KERN_ERR"%s:r > n - 1, err!\n", __func__);
			return -1;
		}

		/*
		 B2：检验s ∈[1,n-1]是否成立，若不成立则验证不通过；
		*/
		BigNum_Convert_Byte_Order_U256(r_s + SM2_BYTE_LEN, work_list->s);
		if (unlikely(BigNum_Check_Zero_U32(work_list->s, SM2_WORD_LEN))) {
			printk(KERN_ERR"%s:s == 0, err!\n", __func__);
			return -1;
		}

		if (unlikely(BigNumCmp(work_list->s, SM2_WORD_LEN, sm2p256v1_params.n, SM2_WORD_LEN) >= 0)) {
			printk(KERN_ERR"%s:s > n - 1, err!\n", __func__);
			return -1;
		}
	}

	ret = __sm2_verify(pke_reg, work_list);
	if (ret) {
		if (unlikely(ret < 0))
			printk(KERN_ERR"%s:__sm2_verify failed!ret:%d\n", __func__, ret);
		return ret;
	}

	return ret;
}

static int sm2_encrypt(volatile pke_reg_t *pke_reg, phytium_sm2_work_list_t *work_list)
{
	struct akcipher_request *req = work_list->req;
	u8 *M, *C;
	int ret;
	u32 outlen;

	if (unlikely(work_list->work_step == 0)) {
		/*
		 e check and convert
		*/
		M = kmalloc(req->src_len + 512, GFP_KERNEL);
		if (unlikely(M == NULL)) {
			printk(KERN_ERR"%s:kmalloc failed!\n", __func__);
			return -ENOMEM;
		}
		C = kmalloc(req->src_len + 512, GFP_KERNEL);
		if (unlikely(C == NULL)) {
			printk(KERN_ERR"%s:kmalloc failed!\n", __func__);
			kfree(M);
			return -ENOMEM;
		}
		sg_copy_buffer(req->src, sg_nents(req->src), M, req->src_len, 0, true);
		work_list->M = M;
		work_list->C = C;
	}

	ret = __sm2_encrypt(work_list->M,
						req->src_len,
						work_list->C,
						&outlen,
						pke_reg,
						work_list);
	if (ret) {
		if (unlikely(ret < 0)) {
			printk(KERN_ERR"%s:__sm2_encrypt failed!ret:%d\n", __func__, ret);
			kfree(work_list->M);
			kfree(work_list->C);
		}
		return ret;
	}

	sm2encrypt_encode(work_list->C, outlen, work_list->M, &outlen);
	if (unlikely(outlen > req->dst_len)) {
		printk(KERN_ERR"%s:outlen[0x%x] > dst_len[0x%x], fail!\n", __func__, outlen, req->dst_len);
		kfree(work_list->M);
		kfree(work_list->C);
		return -1;
	}

	sg_copy_buffer(req->dst, sg_nents(req->dst), work_list->M, outlen, 0, false);
	req->dst_len = outlen;
	kfree(work_list->M);
	kfree(work_list->C);

	return ret;
}

static int sm2_decrypt(volatile pke_reg_t *pke_reg, phytium_sm2_work_list_t *work_list)
{
	struct akcipher_request *req = work_list->req;
	u8 *M, *C;
	int ret;
	u32 outlen;

	if (unlikely(work_list->work_step == 0)) {
		/*
		 e check and convert
		*/
		M = kmalloc(req->src_len + 512, GFP_KERNEL);
		if (unlikely(M == NULL)) {
			printk(KERN_ERR"%s:kmalloc failed!\n", __func__);
			return -ENOMEM;
		}

		C = kmalloc(req->src_len + 512, GFP_KERNEL);
		if (unlikely(C == NULL)) {
			printk(KERN_ERR"%s:kmalloc failed!\n", __func__);
			kfree(M);
			return -ENOMEM;
		}
		sg_copy_buffer(req->src, sg_nents(req->src), M, req->src_len, 0, true);
		ret = sm2decrypt_decode(M, req->src_len, C, &work_list->decode_len);
		if (unlikely(ret)) {
			printk(KERN_ERR"%s:sm2decrypt_decode failed!ret:%d\n", __func__, ret);
			kfree(C);
			kfree(M);
			return ret;
		}
		work_list->M = M;
		work_list->C = C;
	}

	ret = __sm2_decrypt(work_list->C,
						work_list->decode_len,
						work_list->M,
						&outlen,
						pke_reg,
						work_list);
	if (ret) {
		if (unlikely(ret < 0)) {
			printk(KERN_ERR"%s:__sm2_decrypt failed!ret:%d\n", __func__, ret);
			kfree(work_list->M);
			kfree(work_list->C);
		}
		return ret;
	}

	if (unlikely(outlen > req->dst_len)) {
		printk(KERN_ERR"%s:outlen[0x%x] > dst_len[0x%x], fail!\n", __func__, outlen, req->dst_len);
		kfree(work_list->M);
		kfree(work_list->C);
		return -1;
	}

	sg_copy_buffer(req->dst, sg_nents(req->dst), work_list->M, outlen, 0, false);
	req->dst_len = outlen;
	kfree(work_list->M);
	kfree(work_list->C);

	return ret;
}

static int phytium_sm2_work_thread(void *arg)
{
	int i, engine_index, queue_index, work_num, ret;
	phytium_sm2_work_list_t *work_list[4] = {NULL};
	u64 time = jiffies;
	sqlist_t *node;

	queue_index = 0;
	sm3_init(&sm2_rng_ctx);
	while (1) {
		mutex_lock(&scto.pke_lock);
		work_num = 0;

		for (engine_index = 0; engine_index < 1; engine_index++) {
			if (work_list[engine_index] == NULL) {
				for (i = 0; i < NR_CPUS; i++) {
					mutex_lock(&scto.sm2_queue[queue_index].queue_lock);
					node = sqlist_get_head(&scto.sm2_queue[queue_index].list);
					mutex_unlock(&scto.sm2_queue[queue_index].queue_lock);

					queue_index ++;
					if (queue_index >= NR_CPUS)
						queue_index = 0;

					if (node) {
						work_list[engine_index] = container_of(node, phytium_sm2_work_list_t, node);
						break;
					}
				}
			}

			if (work_list[engine_index] == NULL)
				continue;

			work_num ++;
			switch (work_list[engine_index]->work_type) {
				case SM2_SIGN:
					ret = sm2_sign(scto.pke_reg[engine_index], work_list[engine_index]);
					if (ret <= 0) {
						work_list[engine_index]->req->base.complete(&work_list[engine_index]->req->base, ret);
						kfree(work_list[engine_index]);
						work_list[engine_index] = NULL;
					}
					break;
				case SM2_VERIFY:
					ret = sm2_verify(scto.pke_reg[engine_index], work_list[engine_index]);
					if (ret <= 0) {
						work_list[engine_index]->req->base.complete(&work_list[engine_index]->req->base, ret);
						kfree(work_list[engine_index]);
						work_list[engine_index] = NULL;
					}
					break;
				case SM2_ENCRYPT:
					ret = sm2_encrypt(scto.pke_reg[engine_index], work_list[engine_index]);
					if (ret <= 0) {
						work_list[engine_index]->req->base.complete(&work_list[engine_index]->req->base, ret);
						kfree(work_list[engine_index]);
						work_list[engine_index] = NULL;
					}
					break;
				case SM2_DECRYPT:
					ret = sm2_decrypt(scto.pke_reg[engine_index], work_list[engine_index]);
					if (ret <= 0) {
						work_list[engine_index]->req->base.complete(&work_list[engine_index]->req->base, ret);
						kfree(work_list[engine_index]);
						work_list[engine_index] = NULL;
					}
					break;
				default:
					printk(KERN_ERR"work_type[0x%x] ERR!\n", work_list[engine_index]->work_type);
					work_list[engine_index]->req->base.complete(&work_list[engine_index]->req->base, -1);
					kfree(work_list[engine_index]);
					work_list[engine_index] = NULL;
					break;
			}
		}
		mutex_unlock(&scto.pke_lock);

		crypto_rng_update();

		if (work_num == 0) {
			if (running == 0)
				break;
			time = jiffies;
			schedule_timeout_interruptible(1);
		}

		if (time + msecs_to_jiffies(100) < jiffies) {
			time = jiffies;
			schedule();
		}
	}

	running = 1;

	return 0;
}

static int phytium_sm2_insert_queue(struct akcipher_request *req, int work_type)
{
	struct crypto_akcipher *tfm = crypto_akcipher_reqtfm(req);
	phytium_sm2_context_t *ctx = akcipher_tfm_ctx(tfm);
	int cpuid;
	struct scto_dev *dev = ctx->dev;
	phytium_sm2_work_list_t *work_list = kmalloc(sizeof(phytium_sm2_work_list_t), GFP_KERNEL);

	if (unlikely(work_list == NULL)) {
		printk(KERN_ERR"%s:kmalloc failed!\n", __func__);
		return -ENOMEM;
	}

	if (unlikely((req->base.complete == NULL) || (dev == NULL))) {
		printk(KERN_ERR"%s:complete[%p], dev[%p], err!\n", __func__, req->base.complete, dev);
		return -EINVAL;
	}

	if ((work_type == SM2_SIGN) || (work_type == SM2_DECRYPT)) {
		if (unlikely(!ctx->private_key_valid)) {
			printk(KERN_ERR"%s:private_key invalid!\n", __func__);
			return -EINVAL;
		}
	}

	if ((work_type == SM2_VERIFY) || (work_type == SM2_ENCRYPT)) {
		if (unlikely(!ctx->public_key_valid)) {
			printk(KERN_ERR"%s:public_key invalid!\n", __func__);
			return -EINVAL;
		}
	}	

	work_list->work_type = work_type;
	work_list->work_step = 0;
	work_list->req = req;
	work_list->ctx = ctx;

	cpuid = smp_processor_id();
	if (unlikely(cpuid >= NR_CPUS))
		cpuid = 0;

	mutex_lock(&dev->sm2_queue[cpuid].queue_lock);
	sqlist_insert_tail(&dev->sm2_queue[cpuid].list, &work_list->node);
	mutex_unlock(&dev->sm2_queue[cpuid].queue_lock);

	wake_up_process(sm2_task);

	return -EINPROGRESS;
}


static int phytium_sm2_sign(struct akcipher_request *req)
{
	return phytium_sm2_insert_queue(req, SM2_SIGN);
}

static int phytium_sm2_verify(struct akcipher_request *req)
{
	return phytium_sm2_insert_queue(req, SM2_VERIFY);
}

static int phytium_sm2_encrypt(struct akcipher_request *req)
{
	return phytium_sm2_insert_queue(req, SM2_ENCRYPT);
}

static int phytium_sm2_decrypt(struct akcipher_request *req)
{
	return phytium_sm2_insert_queue(req, SM2_DECRYPT);
}

static int phytium_sm2_set_pub_key(struct crypto_akcipher *tfm, const void *key, unsigned int keylen)
{
	phytium_sm2_context_t *ctx = akcipher_tfm_ctx(tfm);
	u32 x[SM2_WORD_LEN], y[SM2_WORD_LEN];

	if (unlikely(key == NULL)) {
		printk(KERN_ERR"pubkey NULL!\n");
		return -1;
	}

	if (unlikely(((u8*)key)[0] != POINT_NOT_COMPRESSED)) {
		printk(KERN_ERR"pubkey only support POINT_NOT_COMPRESSED(0x%x)!\n", POINT_NOT_COMPRESSED);
		return -1;
	}
	if (unlikely(keylen != 1 + 2 * SM2_BYTE_LEN)) {
		printk(KERN_ERR"pubkey keylen:%u no support!\n", keylen);
		return -1;
	}

	BigNum_Convert_Byte_Order_U256(key + 1, x);
	BigNum_Convert_Byte_Order_U256(key + 1 + SM2_BYTE_LEN, y);

	memcpy((u8 *)ctx->x, (u8 *)x, SM2_BYTE_LEN);
	memcpy((u8 *)ctx->y, (u8 *)y, SM2_BYTE_LEN);
	ctx->public_key_valid = 1;

	return 0;
}

static int phytium_sm2_set_priv_key(struct crypto_akcipher *tfm, const void *key, unsigned int keylen)
{
	phytium_sm2_context_t *ctx = akcipher_tfm_ctx(tfm);
	u32 d[SM2_WORD_LEN];

	if (unlikely(key == NULL)) {
		printk(KERN_ERR"prikey NULL!\n");
		return -1;
	}

	if (unlikely(keylen != SM2_BYTE_LEN)) {
	 	printk(KERN_ERR"prikey keylen:%u no support!\n", keylen);
	 	return -1;
	}

	BigNum_Convert_Byte_Order_U256(key, d);

	memcpy((u8 *)ctx->d, (u8 *)d, SM2_BYTE_LEN);
	ctx->private_key_valid = 1;

	return 0;
}

static unsigned int phytium_sm2_max_size(struct crypto_akcipher *tfm)
{
	/* Unlimited max size */
	return PAGE_SIZE;
}

static int phytium_sm2_init(struct crypto_akcipher *tfm)
{
	phytium_sm2_context_t *ctx = akcipher_tfm_ctx(tfm);

	if (unlikely(ctx == NULL)) {
		printk(KERN_ERR"%s:ctx NULL!\n", __func__);
		return -1;
	}

	memset(ctx, 0, sizeof(phytium_sm2_context_t));
	ctx->dev = &scto;

	return 0;
}

static void phytium_sm2_exit(struct crypto_akcipher *tfm)
{
	phytium_sm2_context_t *ctx = akcipher_tfm_ctx(tfm);

	if (unlikely(ctx == NULL)) {
		printk(KERN_ERR"%s:ctx NULL!\n", __func__);
		return;
	}

	memset(ctx, 0, sizeof(phytium_sm2_context_t));
}

static struct akcipher_alg sm2_phytium_alg = {
	.sign = phytium_sm2_sign,
	.verify = phytium_sm2_verify,
	.encrypt = phytium_sm2_encrypt,
	.decrypt = phytium_sm2_decrypt,
	.set_pub_key = phytium_sm2_set_pub_key,
	.set_priv_key = phytium_sm2_set_priv_key,
	.max_size = phytium_sm2_max_size,
	.init = phytium_sm2_init,
	.exit = phytium_sm2_exit,
	.reqsize = 0,
	.base = {
		.cra_name = "sm2",
		.cra_driver_name = "sm2-phytium",
		.cra_priority = 300,
		.cra_module = THIS_MODULE,
		.cra_ctxsize = sizeof(phytium_sm2_context_t),
	},
};

int sm2_phytium_algs_register(void)
{
	int ret;

	running = 1;
	sm2_task = kthread_run(phytium_sm2_work_thread, NULL, "scto_sm2_kthread");
	if (IS_ERR(sm2_task)) {
		printk(KERN_ERR"%s:kthread_run failed!\n", __func__);
		running = 0;
		return -1;
	}

	ret = crypto_register_akcipher(&sm2_phytium_alg);
	if (ret) {
		printk(KERN_ERR"%s:crypto_register_akcipher failed!\n", __func__);
		running = 0;
		while (1) {
			msleep(10);
			if (running)
				break;
		}
		return ret;
	}

	return 0;
}

void sm2_phytium_algs_unregister(void)
{
	crypto_unregister_akcipher(&sm2_phytium_alg);
	running = 0;
	while (1) {
		msleep(10);
		if (running)
			break;
	}
}
#endif
