/* SPDX-License-Identifier: GPL-2.0 */
#ifndef __SM2_PHYTIUM_H__
#define __SM2_PHYTIUM_H__

extern struct scto_dev scto;

#define SM2_MEM_STEP     (32)

//PKE offset
#define PKE_INT_ENABLE_OFFSET     (8)
#define PKE_START_CALC            (1)


/********* pke microcode *********/
#define MICROCODE_PDBL         0x04
#define MICROCODE_PADD         0x08
#define MICROCODE_PVER         0x0C
#define MICROCODE_PMUL         0x10
#define MICROCODE_PMULF        0x14
#define MICROCODE_MODMUL       0x1C
#define MICROCODE_MODINV       0x20
#define MICROCODE_MODADD       0x24
#define MICROCODE_MODSUB       0x28
#define MICROCODE_ADD          0x34
#define MICROCODE_SUB          0x38
#define MICROCODE_CAL_PRE_MON  0x3C
#define MICROCODE_MGMR_PRE_N0  0x40

//some sm2 length
#define SM2_BIT_LEN       (256)
#define SM2_BYTE_LEN      (32)
#define SM2_WORD_LEN      (8)

#define SM2_MAX_ID_BYTE_LEN   (1<<13)

#define POINT_NOT_COMPRESSED      (0x04)

enum {
	SM2_SIGN = 0x20220318,
	SM2_VERIFY,
	SM2_ENCRYPT,
	SM2_DECRYPT,
};


/*pke register struct*/
typedef struct {
	u32 ctrl;
	u32 cfg;
	u32 rev1[2];
	u32 mc_ptr;
	u32 rev2[3];
	u32 stat;
	u32 rt_code;
	u32 rev3[22];
	u32 version;
} pke_reg_t;

typedef struct sm2_domain_parms {
	u32 p[SM2_WORD_LEN];
	u32 p_h[SM2_WORD_LEN];
	u32 a[SM2_WORD_LEN];
	u32 b[SM2_WORD_LEN];
	u32 Gx[SM2_WORD_LEN];
	u32 Gy[SM2_WORD_LEN];
	u32 n[SM2_WORD_LEN];
	u32 n_h[SM2_WORD_LEN];
	u32 n_1[SM2_WORD_LEN];
	u32 Gx_2_128[SM2_WORD_LEN];
	u32 Gy_2_128[SM2_WORD_LEN];
} sm2p256v1_params_t;

typedef struct __sqlist {
	struct __sqlist *next;
	struct __sqlist *prev;
} sqlist_t;

static inline void sqlist_init(sqlist_t *list)
{
	list->next = (sqlist_t *)list;
	list->prev = (sqlist_t *)list;
}

static inline int sqlist_is_empty(sqlist_t *list)
{
	return list->next == list;
}

static inline void sqlist_remove(sqlist_t *node)
{
	node->prev->next = node->next;
	node->next->prev = node->prev;
}

static inline void sqlist_insert_tail(sqlist_t *list, sqlist_t *node)
{
	node->next = list;
	node->prev = list->prev;

	list->prev->next = node;
	list->prev = node;
}


static inline sqlist_t *sqlist_get_head(sqlist_t *list)
{
	sqlist_t *node;

	if(sqlist_is_empty(list))
		return NULL;

	node = list->next;
	sqlist_remove(node);

	return node;
}

typedef struct {
	u32 x[SM2_WORD_LEN];
	u32 y[SM2_WORD_LEN];
	u32 d[SM2_WORD_LEN];
	int public_key_valid;
	int private_key_valid;
	void *dev;
} __attribute__((aligned(64))) phytium_sm2_context_t;

typedef struct {
	sqlist_t node;
	int work_type;
	int work_step;
	struct akcipher_request *req;
	phytium_sm2_context_t *ctx;
	u8 *M;
	u8 *C;
	u32 decode_len;
	u32 e[SM2_WORD_LEN];
	u32 k[SM2_WORD_LEN];
	u32 r[SM2_WORD_LEN];
	u32 s[SM2_WORD_LEN];
	u32 x1[SM2_WORD_LEN];
	u32 y1[SM2_WORD_LEN];
	u32 tmp1[SM2_WORD_LEN];
	u32 tmp2[SM2_WORD_LEN];
	u32 R[SM2_WORD_LEN];
	u32 t[SM2_WORD_LEN];	
} __attribute__((aligned(64))) phytium_sm2_work_list_t;

typedef struct {
	struct mutex queue_lock;
	sqlist_t list;
} __attribute__((aligned(64))) phytium_sm2_queue_t;


/*API*/
int sm2_phytium_algs_register(void);

void sm2_phytium_algs_unregister(void);

#endif
